#! /usr/bin/python -u

#-------------------------------------------------------------------------------
#
# Copyright 2015 Cumulus Networks, inc  all rights reserved
#
# This pass_persist script handles the dot1dBasePortEntry and
# dot1dBasePortIfIndex tables in the BRIDGE-MIB and dot1qBase, dot1qFdbEntry,
# dot1qTpFdbEntry, dot1qTpFdbStatus, and the dot1qVlanStaticName tables in the
# Q-BRIDGE-MIB tables.
#
# To use this script, place the file in
# /usr/share/snmp, change the permissions (chmod 755) and edit
# /etc/snmp/snmpd.conf and add the following two lines:
#
#        view   systemonly  included   .1.3.6.1.2.1.17
#        pass_persist .1.3.6.1.2.1.17 /usr/share/snmp/bridge_pp.py
#
#-------------------------------------------------------------------------------

import subprocess
import getopt
import sys
import os.path
import re
import syslog

import snmp_passpersist as snmp

oid_base = '.1.3.6.1.2.1.17'
# To run from the commandline for debugging,
# simply set debug to non-zero and run to see
# STDOUT output.
debug = 0
syslog_already_logged = False

def pp_add_int(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_int(oid, val)

def pp_add_gau(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_gau(oid, val)

def pp_add_cnt(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_cnt_32bit(oid, val)

def pp_add_str(oid, val):
    if debug:
        print '%s : %s' %(oid, val)
    else:
        pp.add_str(oid, val)

def pp_add_oid(oid, type, val):
    if debug:
        print '%s : %s' %(oid, type, val)
    else:
        pp.add_oid_entry(oid, type, val)

def update ():
    global syslog_already_logged
    # Get interface descriptions from ip link output
    try:
        iplink_lines = subprocess.check_output((['/bin/ip', '-d', '-o', 'link','show']),
                                               stderr=subprocess.STDOUT,
                                               shell=False).splitlines()

        fdb_lines = subprocess.check_output((['/sbin/bridge', 'fdb','show']),
                                            stderr=subprocess.STDOUT,
                                            shell=False).splitlines()
    except Expection as e:
        syslog.syslog('bridge_pp: ethtool EXCEPTION=%s' % e)


    # vlan array is indexed by vlan number
    vlan_dict = {}
    # split all our ip link output to get indexes
    iplink_split = [l.strip().split(' ') for l in iplink_lines]
    # create our if array indexed by name, value is ifindex
    bridge_dict = {}
    lowest_bridge_ifindex = 999999
    lowest_bridge_ifname = None
    # create an interface dict indexed by ifindex string
    if_dict = {}
    for l in iplink_split:
        if len(l) > 1:
            ifindex = l[0].replace(':','')
            ifname = l[1].replace(':','')
            # vlans have an @ symbol, strip it
            if '@' in ifname:
                ifname = ifname.split('@')[0].strip()
            # also keep index by name for vlan use
            if_dict[ifname] = ifindex
            # create empty dict indexed by the ifindex
            if_dict.setdefault(ifindex,{})
            if_dict[ifindex].setdefault('ifname',ifname)
            if 'link/ether' in l:
                machex = l[l.index('link/ether')+1]
                if_dict[ifindex].setdefault('machex',machex)
            if 'bridge' in l:
                # we have a bridge, save the name and index
                bridge_dict.setdefault(ifname,{})
                bridge_dict[ifname].setdefault('ifindex',ifindex)
                # grab the lowest bridge ifindex since that is all
                # the BRIDGE-MIB handles.  We will log to syslog later
                if int(lowest_bridge_ifindex) > int(ifindex):
                    lowest_bridge_ifindex = ifindex
                    lowest_bridge_ifname = ifname
            if 'bridge_slave' in l and 'master' in l:
                # grab the bridge and create the entry if needed
                master = l[l.index('master')+1]
                if_dict[ifindex].setdefault('master',master)
                # since master is a bridge name, just add that
                # bridge slave to a list of slaves
                bridge_dict.setdefault(master,{})
                bridge_dict[master].setdefault('slaves',[])
                bridge_dict[master]['slaves'].append(ifindex)

    #############################################################################
    # each vlan number entry is another array indexed with decimal macs
    fdb_split =  [l.split(' ') for l in fdb_lines]
    for l in fdb_split:
        if len(l) > 6 and l[3]=='vlan':
            perm = False
            if len(l) > 4 and l[-1] == 'permanent':
                # we probably have the word permanent at the end
                perm = True
            vlanid = l[4]
            ifname = l[2]
            if vlanid == '0' and  '.' in ifname:
                # in the non-vlan aware bridge, we may have the vlan in
                # the interface name
                vlanid = ifname.split('.')[1]
            # create empty array if it doesn't exist
            vlan_dict.setdefault(vlanid,{})
            # grab the port index or 0 if there is not any
            ifindex = if_dict.get(ifname,'0')
            # string with decimal representation of mac address
            machex = l[0]
            macdec = '.'.join([str(int(m,16)) for m in machex.split(':')])
            vlan_dict[vlanid].setdefault(machex,{})
            vlan_dict[vlanid][machex].setdefault('ifindex',ifindex)
            vlan_dict[vlanid][machex].setdefault('permanent',perm)
            vlan_dict[vlanid][machex].setdefault('machex',machex)
            vlan_dict[vlanid][machex].setdefault('macdec',macdec)

    # Bridge MIB only handles 1 bridge
    # If we have more then one bridge, simply
    # log to syslog once and use the lowest bridge ifindex
    if len(bridge_dict.keys()) > 1:
        # This MIB only handles one bridge.
        # using the first indexed bridge
        if not syslog_already_logged:
            syslog.syslog("bridge_pp.py: INFO: More then one bridge found. Only using %s." % \
                          lowest_bridge_ifname)
            syslog_already_logged=True
        single_bridge_dict = bridge_dict[lowest_bridge_ifname]
    elif len(bridge_dict.keys()) == 0:
        # yes, we have no bridges at all
        single_bridge_dict = None
    else:
        # we must have exactly 1 bridge, great
        single_bridge_dict = bridge_dict[bridge_dict.keys()[0]]

    # add some anchors for the first part of the bridge mib
    dot1dBase = '1'
    dot1dBaseBridgeAddress = '1'
    dot1dBaseNumPorts = '2'
    dot1dBaseType = '3'
    dot1dBasePortTable = '4'
    dot1dBasePortEntry = '1'
    dot1dBasePort = '1'
    dot1dBasePortIfIndex = '2'

    # add some anchors for the first part of the bridge mib
    pp_add_str('%s'%dot1dBase, '')

    # we may not have a bridge at all
    if single_bridge_dict:
        # there should be only 0 or 1 bridge at this point, grab the mac address
        bridge_index = single_bridge_dict['ifindex']
        macString = ':'.join(map(lambda x: chr(int(x,16)),
                                if_dict[bridge_index]['machex'].split(':')))
        pp_add_str('%s.%s'%(dot1dBase,dot1dBaseBridgeAddress),
                   if_dict[bridge_index]['machex'])
        # just count the number of slaves
        slave_list = single_bridge_dict['slaves']
        pp_add_int('%s.%s'%(dot1dBase,dot1dBaseNumPorts), len(slave_list))
        # we are only a transparent bridge
        pp_add_int('%s.%s'%(dot1dBase,dot1dBaseType), 2)

        #########################################################################
        # handle the 4.1.1 index table
        pp_add_int('%s.%s'%(dot1dBase,dot1dBasePortTable), 0)
        pp_add_int('%s.%s.1'%(dot1dBase,dot1dBasePortTable), 0)
        pp_add_int('%s.%s.1.1'%(dot1dBase,dot1dBasePortTable), 0)

        for i in slave_list:
            pp_add_int('%s.%s.%s.%s.%s' % \
                       (dot1dBase,dot1dBasePortTable,
                        dot1dBasePortEntry,dot1dBasePort,i), i)

        #########################################################################
        # handle the 4.1.2 portifindex table
        pp_add_int('%s.%s.1.2'%(dot1dBase,dot1dBasePortTable), 0)
        for i in slave_list:
            pp_add_int('%s.%s.%s.%s.%s' % \
                       (dot1dBase,dot1dBasePortTable,
                        dot1dBasePortEntry,dot1dBasePortIfIndex,i), i)

    #############################################################################
    # Now we handle some of the Q-BRIDGE-MIB
    # some anchors to get us to qBridgeMIB
    dot1dStp = '2'
    dot1dSr = '3'
    dot1dTp = '4'
    dot1dStatic = '5'
    pBridgeMIB = '6'
    pp_add_str('%s'%dot1dStp, '')
    pp_add_str('%s'%dot1dSr, '')
    pp_add_str('%s'%dot1dTp, '')
    pp_add_str('%s'%dot1dStatic, '')
    pp_add_int('%s'%pBridgeMIB, '')
    qBridgeMIB = '7'
    qBridgeMIBObjects = '1'
    dot1qBase = '1'
    pp_add_str('%s'%qBridgeMIB, '')
    pp_add_str('%s.%s'%(qBridgeMIB,qBridgeMIBObjects), '')
    pp_add_str('%s.%s.%s'%(qBridgeMIB,qBridgeMIBObjects,dot1qBase), '')
    #############################################################################
    # handle the 1q base table
    dot1qBase_oid = '7.1.1'
    dot1qVlanVersionNumber = '1'
    dot1qMaxVlanId = '2'
    dot1qMaxSupportedVlans = '3'
    dot1qNumVlans = '4'
    dot1qGvrpStatus = '5'
    numVlans = str(len(vlan_dict.keys()))
    pp_add_int('%s.%s.0' % (dot1qBase_oid, dot1qVlanVersionNumber), '1')
    pp_add_int('%s.%s.0' % (dot1qBase_oid, dot1qMaxVlanId), '4094')
    pp_add_gau('%s.%s.0' % (dot1qBase_oid, dot1qMaxSupportedVlans), 4094)
    # since linux always shows 0 as a vlan, we need to count that too
    pp_add_gau('%s.%s.0' % (dot1qBase_oid, dot1qNumVlans), numVlans)
    # since we are not doing gvrp, we say it is disabled value is 2
    pp_add_int('%s.%s.0' % (dot1qBase_oid, dot1qGvrpStatus), '2')

    #############################################################################
    # grab all vlans and sort them as integers
    vlans = [int(i) for i in vlan_dict.keys()]
    vlans.sort()

    #############################################################################
    pp_add_int('7.1.2', '0')
    pp_add_int('7.1.2.1', '0')
    pp_add_int('7.1.2.1.1', '0')
    pp_add_cnt('7.1.2.1.1.2', '0')
    dot1qFdbEntry = '7.1.2.1.1'
    dot1qFdbDynamicCount = '2'
    # show the dynamic count for each vlan
    for vlan in vlans:
        myindex = vlans.index(vlan) + 1
        # count the number of macs for each vlan
        entryCount = len(vlan_dict[str(vlan)].keys())  #['macport_list_dec'])
        pp_add_cnt('%s.%s.%s' % (dot1qFdbEntry, dot1qFdbDynamicCount,
                                 myindex), entryCount)

    #############################################################################
    pp_add_int('7.1.2.2', '0')
    pp_add_int('7.1.2.2.1', '0')
    pp_add_int('7.1.2.2.1.2', '0')
    dot1qTpFdbEntry = '7.1.2.2.1'
    dot1qTpFdbPort = '2'
    # this is an important table that lists the vlans, macs, and the port index
    for vlan in vlans:
        myindex = vlans.index(vlan) + 1
        # just grab the decimal dotted macs we made earlier
        machexs = vlan_dict[str(vlan)].keys()
        machexs.sort()
        for mac in machexs:
            macdec = vlan_dict[str(vlan)][mac]['macdec']
            port = vlan_dict[str(vlan)][mac]['ifindex']
            pp_add_int('%s.%s.%s.%s' % (dot1qTpFdbEntry, dot1qTpFdbPort,
                                        myindex, macdec), port)

    #############################################################################
    pp_add_int('7.1.2.2.1.3', '0')
    dot1qTpFdbEntry = '7.1.2.2.1'
    dot1qTpFdbStatus = '3'
    # this table lists the vlans, macs, and the port status
    # need to make sure the mac address are sorted in hex lexical
    # order since the dotted decimal does not sort correctly.
    for vlan in vlans:
        myindex = vlans.index(vlan) + 1
        # just grab the decimal dotted macs we made earlier
        machexs = vlan_dict[str(vlan)].keys()
        machexs.sort()
        for mac in machexs:
            permanent = vlan_dict[str(vlan)][mac]['permanent']
            macdec = vlan_dict[str(vlan)][mac]['macdec']
            if permanent:
                status = 4 # 4 means self or our own mac address
            else:
                status = 3 # learned
            pp_add_int('%s.%s.%s.%s' % (dot1qTpFdbEntry, dot1qTpFdbStatus,
                                        myindex, macdec), status)

    #############################################################################
    pp_add_int('7.1.4', '0')
    pp_add_int('7.1.4.3', '0')
    pp_add_int('7.1.4.3.1', '0')
    pp_add_str('7.1.4.3.1.1', '0')
    dot1qVlanStaticEntry = '7.1.4.3.1'
    dot1qVlanStaticName = '1'
    # show the vlans with names, and show the default 0 vlan too
    for vlan in vlans:
        myindex = vlans.index(vlan) + 1
        # just make up a VLAN name like VLAN-1234 or default
        if myindex == 1:
            vlanName = 'default'
        else:
            vlanName = 'VLAN-%s' % vlan
        pp_add_str('%s.%s.%s' % (dot1qVlanStaticEntry, dot1qVlanStaticName,
                                 myindex), vlanName)

    #############################################################################

if debug:
    update()
else:
    pp = snmp.PassPersist(oid_base)
    pp.start(update, 60)
