#! /usr/bin/python -u
# Copyright 2016 Cumulus Networks LLC, all rights reserved
#
# This is an implementation of the Entity MIB (RFC 4133)
# which is mostly the output of the decode-syseeprom command.
# To activate this script, snmpd must be running and this
# script should be installed /usr/share/snmp/entity_pp.py
# and include the following line in /etc/snmp/snmpd.conf
#
#   pass_persist .1.3.6.1.2.1.47  /usr/share/snmp/entity_pp.py
#

import subprocess
import sys
import json
import time
import os
import syslog
import snmp_passpersist as snmp

debug = 0
entityMIB = '.1.3.6.1.2.1.47'

entPhysicalTable = '1.1.1'
entPhysicalEntry = '1.1.1.1'

# we only handle the PhysicalTable for now
entLogicalTable  = '1.2.1'
entAliasMappingIdentifier = '1.3.3.1'
# we will just set this to uptime since the physical table is static
entLastChangeTime = '1.4.1'

# grab the uptime timetics
try:
    fd = open('/proc/uptime', 'r')
    uptime = int(float(fd.read().split()[0]) * 100)
    fd.close()
except:
    uptime = 0

# isFru values
fru_true = 1
fru_false = 2

# grab the software version
try:
    osrelease = [l.split('=') for l in open('/etc/os-release','r').read().split('\n')]
    linuxversion = [l.split() for l in open('/proc/version','r').read().split('\n')]
    # remove quotes from cumulus linux
    clVersion = '%s %s (%s %s)' % (osrelease[0][1][1:-1],
                                   osrelease[1][1],
                                   osrelease[2][1][1:-1],
                                   linuxversion[0][2])
except:
    clVersion = ''

try:
    # easy way to get a list of physical interfaces
    ports = [i.split()[0] for i in open('/var/lib/cumulus/porttab','r').\
             readlines() if not i.startswith('#') and i.startswith('swp')]
except:
    # if we have a problem, just create an empty list
    ports = []

# ignoring mfgDate for now as the snmp_passpersist module cannot handle
# dateAndTime attributes.
# grab the decode-syseeprom for hardware rev, serial number
# we really should be using one interface and that is dmidecode.
hardwareRev = ''
serialNum = ''
mfgDate = ''
modelName = ''
assetID = ''
mfgName = ''

classes = {'chassis':3, 'container':5, 'powerSupply':6,
           'fan':7, 'sensor':8, 'port':10}

# Grab the sensor data first
# this is just a list of dictionaries, one for each sensor
# Since we only really need the description here, we only need to
# call this once at the beginning.
try:
    smonctlJsonList = json.loads(subprocess.check_output((['/usr/sbin/smonctl', '-j']),
                                                         stderr=subprocess.STDOUT,
                                                         shell=False))
except:
    smonctlJsonList = []

if debug:
    print('entity_pp.py: running in debug mode')

logicalTableRows = ['index', 'description', 'type', 'comminity',
                    'tAddress', 'tDomain', 'contextEngineID', 'contextName']
# physical classes
classes = {'chassis':'3',
           'container':'5',
           'power':'6',
           'fan':'7',
           'temp':'8',
           'port':'10'}

entityIDStart = {'chassis'  :        0,
                 'container':       10,
                 'temp'     :100000000,
                 'fan'      :100011000,
                 'power'    :110000000,
                 'port'     :111000000,
                 }

entityNames = entityIDStart.keys()

physicalTableRows = ['index', 'description', 'vendorType', 'containedIn',
                     'class', 'parentRelPos', 'name', 'hardwareRev',
                     'firmwareRev', 'softwareRev', 'serialNum', 'mfgName',
                     'modelName', 'alias', 'assetID', 'isFRU', 'mfgDate',
                     'uris']
# these are the elements of the table we will loop over
# most vendors seem to skip the first item called index so we skip it too
entPhysicalTable = list(enumerate(physicalTableRows, start=1))

# we will not show the mfgDate until we can support dateAndTime objects
physicalTableDrops = ['mfgDate']

oidType = {'index':'add_str',
           'description':'add_str',
           'vendorType':'add_oid',
           'containedIn':'add_int',
           'class':'add_int',
           'parentRelPos':'add_int',
           'name':'add_str',
           'hardwareRev':'add_str',
           'firmwareRev':'add_str',
           'softwareRev':'add_str',
           'serialNum':'add_str',
           'mfgName':'add_str',
           'modelName':'add_str',
           'alias':'add_str',
           'assetID':'add_str',
           'isFRU':'add_int',
           'mfgDate':'add_str',
           'uris':'add_str'
           }

rowDefault = {'index':'',
              'description':'',
              'vendorType':'0.0',
              'containedIn':1,
              'class':0,
              'parentRelPos':1,
              'name':'',
              'hardwareRev':'',
              'firmwareRev':'',
              'softwareRev':'',
              'serialNum':'',
              'mfgName':'',
              'modelName':'',
              'alias':'',
              'assetID':'',
              'isFRU':fru_false,
              'mfgDate':'0-0-0,0:0:0.0,.0:0',
              'uris':''
              }

# there is only one chassis, just define it, leave defaults blank
# FIXME we will need to fix these when we get chassis info
chassisList = [{'description':'',  # FIXME
                'class':classes['chassis'],
                'name':'',   # FIXME
                'hardwareRev':hardwareRev,
                'firmwareRev':clVersion,
                'softwareRev':clVersion,
                'serialNum':serialNum,
                'modelName':modelName,
                'mfgName':mfgName,
                'type':'chassis'
                }]

# create the port list of arrays, initializing the type and description
#  FIXME we will need to fix these when we get more port info
portList = [{'description':p,
             'name':p,
             'class':classes['port'],
             'mfgName':'', # FIXME LATER
             'modelName':'', # FIXME LATER
             'type':'port'}
            for p in ports]

temperatureList = [s for s in smonctlJsonList if s['type'] == 'temp']
fanList = [s for s in smonctlJsonList if s['type'] == 'fan']
powerList = [s for s in smonctlJsonList if s['type'] == 'power']

# and one list to bind them all
# until we get reliable chassis and port information (CM-8955),
# we do not include chassis data
entityGroupList = [temperatureList, fanList, powerList, portList]

# add the ID for each device
for entityList in entityGroupList:
    IDCounter = 0
    for entity in entityList:
        IDCounter += 1
        entity['ID'] = entityIDStart[entity['type']] + IDCounter
        entity['class'] = classes[entity['type']]

def update():
    """
    Nothing really gets updated since the physical entities do not
    change.  Until we start supporting hot swapping of chassis, or
    networking card, we are static.
    """
    # we start with description rows, which is second
    for rowval,rowname in entPhysicalTable[1:]:
        # loop over the entity groups like
        # chassis, ports, temp sensors, fan sensors, power sensors
        # some rows will not get used so check the drop list
        if rowname in physicalTableDrops:
            # e.g. mfgDate
            continue
        for entityGroup in entityGroupList:
            # finally, each group has a list of actual entities
            for entity in entityGroup:
                if debug:
                    myfunc = oidType[rowname]
                    myval = entity.get(rowname,rowDefault[rowname])
                    print('%s.%s.%s' % (entPhysicalEntry, rowval,
                                        entity['ID']), myval)
                else:
                  try:
                    myfunc = getattr(pp, oidType[rowname])
                    myval = entity.get(rowname,rowDefault[rowname])
                    # here, we will call add_int, add_str, add_tt, or add_gau
                    myfunc('%s.%s.%s' % (entPhysicalEntry, rowval,
                                         entity['ID']), myval)
                  except Exception as e:
                      syslog.syslog(" ERROR: %s" % e)

    # handle the entLastChange
    if debug:
        print('%s.%s' % (entLastChangeTime, 0), uptime)
    else:
        pp.add_tt('%s.%s' % (entLastChangeTime, 0), uptime)


syslog.syslog("entity_pp.py handling ENTITY MIB: starting...")
if debug:
    update()
else:
    pp = snmp.PassPersist(entityMIB)
    syslog.syslog("entity_pp.py: initialized...")
    pp.start(update, 31536000) # update once a year

